"""
The module runs zserio compiler and provides convenient method to generate Python sources.
"""

import os
import subprocess
import sys
import shutil
import importlib
import typing

class JavaNotFoundException(Exception):
    """
    Exception thrown if Java executable has been not found.
    """

def runCompiler(cmdArgs: typing.List[str], *, captureOutput = True,
                checkExitCode = False) -> subprocess.CompletedProcess:
    """
    Runs zserio compiler using jar file stored in the zserio pip package.

    :param cmdArgs: List of strings which represents zserio command line arguments.
    :param captureOutput: True to capture stdout and stderr.
    :param checkExitCode: True to raise a subprocess.CalledProcessError exception if the zserio compiler fails.
    :returns: subprocess.CompletedProcess instance returned from subprocess.run() method.
    :raises JavaNotFoundException: If Java is not found.
    """

    javaExectuable = _findJavaExecutable()
    zserioCommand = [javaExectuable, "-jar", ZSERIO_JAR_FILE]
    zserioCommand[len(zserioCommand):] = cmdArgs

    return subprocess.run(zserioCommand, capture_output = captureOutput, check = checkExitCode, text = True)

def generatePython(mainZsFile: str, *, isDefaultPackage: bool = False, zsDir: str = None, genDir: str = None,
                   topLevelPackage: str = None, extraArgs: typing.List[str] = None) -> typing.Any:
    """
    Generates Python sources by running zserio compiler.

    The generated Python package API will be automatically imported and returned as a result.

    The generated Python package will be as well automatically added to the pythonpath.

    If a directory where to generate Python sources is not specified, Python sources will be generated into
    the default directory `.zserio_python_package`. This default directory is created either in specified zserio
    source file directory or in directory where main zserio source file is located.

    Example using implicit imported API:
        ```
        applApi = generatePython("test/structure.zs", zsDir = "zs", genDir = "gen", topLevelPackage = "appl")
        testStructure = applApi.test.structure.TestStructure()
        ```

    Example using explicit import:
        ```
        generatePython("test/structure.zs", zsDir = "zs", genDir = "gen", topLevelPackage = "appl")
        structureModule = importlib.import_module("appl.test.structure.TestStructure")
        testStructure = structureModule.TestStructure()
        ```

    :param mainZsFile: Main zserio source file to compile.
    :param isDefaultPackage: True if Zserio source file is default package.
    :param zsDir: Zserio source file directory ('-src' command line option).
    :param genDir: Directory where to generate Python sources ('-python <dir>' command line option).
    :param topLevelPackage: Top level package for compilation ('-setTopLevelPackage <pkg>' command line option).
    :param extraArgs: List of extra command line options.
    :returns: Imported api module of generated Python sources.
    :raises: JavaNotFoundException: If Java is not found.
    :raises: subprocess.CalledProcessError: If calling zserio compiler process failed.
    """

    cmdArgs = [mainZsFile]
    if zsDir is not None:
        cmdArgs += ["-src", zsDir]

    if genDir is not None:
        pythonDir = genDir
    elif zsDir is not None:
        pythonDir = os.path.join(zsDir, ZSERIO_DEFAULT_GEN_DIR_NAME)
    else:
        pythonDir = os.path.join(os.path.dirname(os.path.abspath(mainZsFile)), ZSERIO_DEFAULT_GEN_DIR_NAME)
    cmdArgs += ["-python", pythonDir]

    if topLevelPackage is not None:
        cmdArgs += ["-setTopLevelPackage", topLevelPackage]
    if extraArgs is not None:
        cmdArgs += extraArgs

    runCompiler(cmdArgs, checkExitCode = True)

    return _importApiModule(mainZsFile, isDefaultPackage, pythonDir, topLevelPackage)

def generate(srcFile: str = "", moduleName: str = "") -> None:
    """
    Generates Python sources by running zserio compiler given main zserio source with full path.

    The generated Python package will be added automatically to the pythonpath.

    This method is DEPRECATED and will be removed in the next release! Please note, that this method
    works correctly only if main zserio source package contains only one package id (see example below)!

    Example of main zserio file `structure.zs` which does not compile by this method:
        ```
        package test.structure;

        struct TestStructure
        {
            int32 value;
        };
        ```

    The `moduleName` argument allows to set a top-level python module name,
    under which the generated sources will be placed.

    Examples:

        With top-level package:
            ```
            import zserio
            zserio.generate("myfile.zs", "mypackage")
            from mypackage.myfile import *
            ```

        Without top-level package:
            ```
            import zserio
            zserio.generate("myfile.zs")
            from myfile import *
            ```

    :param srcFile: Source zserio file.
    :param moduleName: (Optional) Top-level package directory name.
    """

    mainZsFile = os.path.basename(srcFile)
    zsDir = os.path.dirname(os.path.abspath(srcFile))

    generatePython(mainZsFile, zsDir = zsDir, topLevelPackage = moduleName)

def _findJavaExecutable() -> str:
    javaHome = os.getenv("JAVA_HOME", None)
    if javaHome:
        javaPath = os.path.join(javaHome, "bin")
        javaExecutable = shutil.which("java", path=javaPath)
        if not javaExecutable:
            raise JavaNotFoundException("compiler: Java not found (wrong ${JAVA_HOME})")
    else:
        javaExecutable = shutil.which("java")
        if not javaExecutable:
            raise JavaNotFoundException("compiler: Java not found (checked ${JAVA_HOME} and ${PATH})")

    return javaExecutable

def _importApiModule(mainZsFile: str, isDefaultPackage: bool, pythonDir: str,
                     topLevelPackage: str = None) -> typing.Any:
    absPythonDir = os.path.abspath(pythonDir)
    sys.path.append(absPythonDir)

    apiModulePath = "api"
    if not isDefaultPackage:
        # we need to find out the first left most part of path
        if topLevelPackage is not None:
            apiModulePathPrefix = topLevelPackage.split(".")[0]
        else:
            mainZsWithoutExt = os.path.splitext(mainZsFile)[0]
            apiModulePathPrefix = mainZsWithoutExt.split(os.sep)[0]

        apiModulePath = apiModulePathPrefix + "." + apiModulePath

    return importlib.import_module(apiModulePath)

ZSERIO_JAR_FILE = os.path.join(os.path.dirname(os.path.abspath(__file__)), "compiler", "zserio.jar")
ZSERIO_DEFAULT_GEN_DIR_NAME = ".zserio_python_package"



"""
Main entry point of zserio pip module.
"""

import sys

import zserio.compiler

def main() -> int:
    """
    Main entry point of zserio pip module.

    This method envokes zserio compilers. It is called if zserio pip module is called on the command line
    (using 'python3 -m zserio').

    :returns: Exit value of zserio compiler.
    """
    completedProcess = zserio.runCompiler(sys.argv[1:], captureOutput = False)
    sys.exit(completedProcess.returncode)

if __name__ == "__main__":
    main()



import unittest
import os
import distutils.dir_util
import importlib
import subprocess

from zserio.compiler import JavaNotFoundException, runCompiler, generatePython, generate

class CompilerTest(unittest.TestCase):

    def testRunCompiler(self):
        completedProcess = runCompiler(["--help"])
        self.assertEqual(0, completedProcess.returncode)

        completedProcess = runCompiler([])
        self.assertEqual(1, completedProcess.returncode)

    def testJavaHomeNotFound(self):
        javaHomeExists = "JAVA_HOME" in os.environ
        if javaHomeExists:
            javaHome = os.environ["JAVA_HOME"]
        os.environ["JAVA_HOME"] = "wrong_path"
        with self.assertRaises(JavaNotFoundException):
            runCompiler(["--help"])
        if javaHomeExists:
            os.environ["JAVA_HOME"] = javaHome
        else:
            del os.environ["JAVA_HOME"]

    def testJavaOnPathNotFound(self):
        javaHomeExists = "JAVA_HOME" in os.environ
        if javaHomeExists:
            javaHome = os.environ["JAVA_HOME"]
            del os.environ["JAVA_HOME"]
        path = os.environ["PATH"]
        os.environ["PATH"] = ""
        with self.assertRaises(JavaNotFoundException):
            runCompiler(["--help"])
        os.environ["PATH"] = path
        if javaHomeExists:
            os.environ["JAVA_HOME"] = javaHome

    def testGenerateCompilationFailure(self):
        with self.assertRaises(subprocess.CalledProcessError):
            generatePython("invalid_source.zs")

    def testGeneratePythonDefaultPackage(self):
        zsDir = os.path.join(self._getTestZsDir(), "default_package")
        buildDir = os.path.join(self._getBuildTestDir(), "generate_python_default_package")
        testZsDir = os.path.join(buildDir, "zs")
        distutils.dir_util.copy_tree(zsDir, testZsDir)
        genDir = os.path.join(buildDir, "gen")
        mainZsFile = "structure_default.zs"
        currentDir = os.getcwd()
        os.chdir(testZsDir)
        api = generatePython(mainZsFile, isDefaultPackage = True, genDir = genDir)
        os.chdir(currentDir)

        testStructure = api.TestStructure()
        self.assertEqual(0, testStructure.getValue())

        testStructureModule = importlib.import_module("TestStructure")
        importedTestStructure = testStructureModule.TestStructure()
        self.assertEqual(0, importedTestStructure.getValue())

    def testGeneratePythonMainZsWithPath(self):
        zsDir = os.path.join(self._getTestZsDir(), "main_zs_with_path")
        buildDir = os.path.join(self._getBuildTestDir(), "generate_python_main_zs_with_path")
        genDir = os.path.join(buildDir, "gen")
        mainZsFile = os.path.join("company", "main", "structure_with_path.zs")
        companyApi = generatePython(mainZsFile, zsDir = zsDir, genDir = genDir)

        testStructure = companyApi.main.structure_with_path.TestStructure()
        self.assertEqual(0, testStructure.getValue())

        testStructureModule = importlib.import_module("company.main.structure_with_path.TestStructure")
        importedTestStructure = testStructureModule.TestStructure()
        self.assertEqual(0, importedTestStructure.getValue())

    def testGeneratePythonMainZsWithoutPath(self):
        zsDir = os.path.join(self._getTestZsDir(), "main_zs_without_path")
        buildDir = os.path.join(self._getBuildTestDir(), "generate_python_main_zs_without_path")
        genDir = os.path.join(buildDir, "gen")
        mainZsFile = "structure_without_path.zs"
        api = generatePython(mainZsFile, zsDir = zsDir, genDir = genDir)

        testStructure = api.TestStructure()
        self.assertEqual(0, testStructure.getValue())

        structureModule = importlib.import_module("structure_without_path.TestStructure")
        importedTestStructure = structureModule.TestStructure()
        self.assertEqual(0, importedTestStructure.getValue())

    def testGeneratePythonTopLevelPackage(self):
        zsDir = os.path.join(self._getTestZsDir(), "main_zs_without_path")
        buildDir = os.path.join(self._getBuildTestDir(), "generate_python_top_level_package")
        genDir = os.path.join(buildDir, "gen")
        mainZsFile = "structure_without_path.zs"
        # "top_level_package.main" must be unique not to mix paths in python system path
        topLevelPackageApi = generatePython(mainZsFile, zsDir = zsDir, genDir = genDir,
                                            topLevelPackage = "top_level_package.main",
                                            extraArgs = ["-withoutPubsubCode", "-withoutServiceCode"])

        testStructure = topLevelPackageApi.main.structure_without_path.TestStructure()
        self.assertEqual(0, testStructure.getValue())

        structureModule = importlib.import_module("top_level_package.main.structure_without_path.TestStructure")
        importedTestStructure = structureModule.TestStructure()
        self.assertEqual(0, importedTestStructure.getValue())

    def testGeneratePythonWithoutGenDir(self):
        zsDir = os.path.join(self._getTestZsDir(), "main_zs_without_path")
        buildDir = os.path.join(self._getBuildTestDir(), "generate_python_without_gen_dir")
        testZsDir = os.path.join(buildDir, "zs")
        distutils.dir_util.copy_tree(zsDir, testZsDir)
        mainZsFile = "structure_without_path.zs"
        currentDir = os.getcwd()
        os.chdir(testZsDir)
        # "without_gen_dir.main" must be unique not to mix paths in python system path
        withoutGenDirApi = generatePython(mainZsFile, topLevelPackage = "without_gen_dir.main")
        os.chdir(currentDir)

        testStructure = withoutGenDirApi.main.structure_without_path.TestStructure()
        self.assertEqual(0, testStructure.getValue())

        testStructureModule = importlib.import_module(
            "without_gen_dir.main.structure_without_path.TestStructure")
        importedTestStructure = testStructureModule.TestStructure()
        self.assertEqual(0, importedTestStructure.getValue())

    def testGenerate(self):
        zsDir = os.path.join(self._getTestZsDir(), "main_zs_without_path")
        buildDir = os.path.join(self._getBuildTestDir(), "generate")
        testZsDir = os.path.join(buildDir, "zs")
        distutils.dir_util.copy_tree(zsDir, testZsDir)
        zsFile = os.path.join(testZsDir, "structure_without_path.zs")
        generate(zsFile, "generate")

        testStructureModule = importlib.import_module("generate.structure_without_path.TestStructure")
        testStructure = testStructureModule.TestStructure()
        self.assertEqual(0, testStructure.getValue())

    @staticmethod
    def _getBuildTestDir():
        testDir = os.path.dirname(os.path.abspath(__file__))
        testZsDir = os.path.join(testDir, "..", "build", "test")

        return testZsDir

    @staticmethod
    def _getTestZsDir():
        testDir = os.path.dirname(os.path.abspath(__file__))
        zsDir = os.path.join(testDir, "zs")

        return zsDir



import setuptools
import json
import urllib.request
import zipfile
import io
import os
import shutil

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
DEFAULT_BUILD_DIR = os.path.join(ROOT_DIR, "build")
BUILD_DIR = os.getenv("PYPI_BUILD_DIR", DEFAULT_BUILD_DIR)
DOWNLOAD_DIR = os.path.join(BUILD_DIR, "download")

def _downloadLatestZserioRelease() -> str:
    """
    Downloads the latest Zserio release from GitHub.

    The method extracts downloaded zip files to the DOWNLOAD_DIR as well.

    :returns: The latest Zserio release version in string format.
    """
    print("downloading the latest zserio release JSON file", end = "")
    zserio_release_url = urllib.request.urlopen("https://api.github.com/repos/ndsev/zserio/releases/latest")
    zserio_release_json = json.loads(zserio_release_url.read().decode('utf-8'))
    zserio_version = zserio_release_json["tag_name"][1:]
    zserio_bin_zip_url = zserio_release_json["assets"][0]["browser_download_url"]
    zserio_runtime_libs_zip_url = zserio_release_json["assets"][1]["browser_download_url"]
    print(" (found zserio version " + zserio_version + ")")

    print("downloading the latest zserio binaries")
    zserio_bin_zip = urllib.request.urlopen(zserio_bin_zip_url)
    print("extracting the latest zserio binaries")
    zserio_bin_zip_file = zipfile.ZipFile(io.BytesIO(zserio_bin_zip.read()), 'r')
    zserio_bin_zip_file.extractall(DOWNLOAD_DIR)
 
    print("downloading the latest zserio runtime")
    zserio_runtime_libs_zip = urllib.request.urlopen(zserio_runtime_libs_zip_url)
    print("extracting the latest zserio runtime")
    zserio_runtime_libs_zip_file = zipfile.ZipFile(io.BytesIO(zserio_runtime_libs_zip.read()), 'r')
    zserio_runtime_libs_zip_file.extractall(DOWNLOAD_DIR)

    return zserio_version

def _createZserioPyPiPackage():
    """
    Creates Zserio PyPi package.

    Zserio PyPi package is a merge of Zserio Python runtime library and PyPi source directory.

    :returns: The directory where Zserio PyPi package has been created.
    """
    print("copying zserio python runtime and compiler")
    downloaded_runtime_dir = os.path.join(DOWNLOAD_DIR, "runtime_libs", "python", "zserio")
    zserio_package_dir = os.path.join(BUILD_DIR, "zserio")
    shutil.copytree(downloaded_runtime_dir, zserio_package_dir, dirs_exist_ok = True)
    runtime_compiler_dir = os.path.join(zserio_package_dir, "compiler")
    if not os.path.exists(runtime_compiler_dir):
        os.makedirs(runtime_compiler_dir)
    shutil.copyfile(os.path.join(DOWNLOAD_DIR, "zserio.jar"), os.path.join(runtime_compiler_dir, "zserio.jar"))

    pypi_src_dir = os.path.join(ROOT_DIR, "src", "zserio")
    shutil.copytree(pypi_src_dir, zserio_package_dir, dirs_exist_ok = True,
                    ignore = shutil.ignore_patterns("__init__.py"))

    print("extending zserio runtime __init__.py")
    runtime_init_py_file_name = os.path.join(zserio_package_dir, "__init__.py")
    with open(runtime_init_py_file_name, "a+", encoding="utf-8") as runtime_init_py_file:
        pypi_init_py_file_name = os.path.join(pypi_src_dir, "__init__.py")
        with open(pypi_init_py_file_name, "r", encoding="utf-8") as pypi_init_py_file:
            runtime_init_py_file.write("\n")
            runtime_init_py_file.write(pypi_init_py_file.read())

    return BUILD_DIR

def _createPyPiLongDescription() -> str:
    """
    Creates long description for PyPi package from project's README.md file.

    :returns: The PyPi long description.
    """
    read_me_file_name = os.path.join(ROOT_DIR, "README.md")
    with open(read_me_file_name, "r", encoding="utf-8") as file:
        read_me = file.read()
    startIndex = read_me.find("Zserio PyPi package contains")
    if startIndex == -1:
        startIndex = 0
    endIndex = read_me.find("\n## Building")
    if endIndex == -1:
        endIndex = len(read_me)
    long_description = read_me[startIndex:endIndex]

    return long_description

setuptools.setup(
    name="zserio",
    version=_downloadLatestZserioRelease(),
    url="https://github.com/ndsev/zserio-pypi",
    author="Navigation Data Standard e.V.",
    author_email="support@nds-association.org",

    description="Zserio runtime with compiler.",
    long_description=_createPyPiLongDescription(),
    long_description_content_type="text/markdown",

    package_dir={
        '': _createZserioPyPiPackage()
    },
    packages=['zserio'],
    package_data={
        'zserio': ['compiler/zserio.jar', 'py.typed']
    },

    entry_points={
        'console_scripts': ['zserio=zserio.__main__:main']
    },

    python_requires='>=3.8',

    license = "BSD-3 Clause",
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: BSD License"
     ],
)
